/*
 * Copyright (C) 2021-2022  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/notify_object_events.hh"

#include "src/impl/exception.hh"
#include "src/impl/strong_type.hh"
#include "src/impl/unwrap_uuid_from_psql.hh"

#include "src/util/util.hh"

#include "libfred/notifier/gather_email_data/gather_email_addresses.hh"
#include "libfred/notifier/gather_email_data/gather_email_content.hh"
#include "libfred/object/registry_object_type.hh"

#include "libhermes/libhermes.hh"

#include "liblog/liblog.hh"

#include "libpg/pg_rw_transaction.hh"
#include "libpg/query.hh"

#include "libstrong/type.hh"

#include <boost/algorithm/string/trim.hpp>
#include <boost/uuid/string_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <vector>

namespace Fred {
namespace Notify {
namespace ObjectEvents {
namespace Impl {

namespace {

std::string to_string(Notification::ObjectType _value)
{
    switch (_value)
    {
        case Notification::ObjectType::contact: return "contact";
        case Notification::ObjectType::nsset: return "nsset";
        case Notification::ObjectType::domain: return "domain";
        case Notification::ObjectType::keyset: return "keyset";
    }
    throw std::logic_error("value doesn't exist in Notification::ObjectType");
}

std::string to_string(Notification::NotifiedEvent _value)
{
    switch (_value)
    {
        case Notification::NotifiedEvent::created: return "created";
        case Notification::NotifiedEvent::updated: return "updated";
        case Notification::NotifiedEvent::transferred: return "transferred";
        case Notification::NotifiedEvent::deleted: return "deleted";
        case Notification::NotifiedEvent::renewed: return "renewed";
    }
    throw std::logic_error("value doesn't exist in Notification::ObjectState");
}

/**
 * Helps to implement from_db_handle function using to_db_handle conversion function.
 * @tparam S type of source value
 * @tparam D type of destination value
 * @tparam items number of all possible destination values
 * @param src source value
 * @param set_of_results C array of all possible destination values
 * @param forward_transformation function which transforms D value into S value
 * @return matching counterpart value
 * @throw std::invalid_argument in case that conversion fails
 */
template <class S, class D, unsigned items>
D inverse_transformation(const S& src, const D (&set_of_results)[items], S (*forward_transformation)(D))
{
    const D* const end_of_results = set_of_results + items;
    for (const D* result_candidate_ptr = set_of_results; result_candidate_ptr < end_of_results; ++result_candidate_ptr)
    {
        if (forward_transformation(*result_candidate_ptr) == src)
        {
            return *result_candidate_ptr;
        }
    }
    throw std::invalid_argument("conversion does not exist");
}

template <typename T>
T from_string(const std::string& _str);

template <>
Notification::ObjectType from_string<Notification::ObjectType>(const std::string& _str)
{
    constexpr Notification::ObjectType possible_results[] =
            {
                Notification::ObjectType::contact,
                Notification::ObjectType::nsset,
                Notification::ObjectType::domain,
                Notification::ObjectType::keyset
            };
    return inverse_transformation(_str, possible_results, to_string);
}

template <>
Notification::NotifiedEvent from_string<Notification::NotifiedEvent>(const std::string& _str)
{
    constexpr Notification::NotifiedEvent possible_results[] =
            {
                Notification::NotifiedEvent::created,
                Notification::NotifiedEvent::updated,
                Notification::NotifiedEvent::transferred,
                Notification::NotifiedEvent::deleted,
                Notification::NotifiedEvent::renewed
            };
    return inverse_transformation(_str, possible_results, to_string);
}

using ContactId = LibStrong::Type<int, struct ContactIdTag_>;
using RegistrarId = LibStrong::Type<unsigned long long, struct RegistrarIdTag_>;
using HistoryIdPostChange = LibStrong::Type<unsigned long long, struct HistoryIdPostChangeTag_>;
using SvtrId = StrongString<struct SvtrIdTag_>;
using ObjectUuid = LibStrong::Type<boost::uuids::uuid, struct ObjectUuidTag_>;
using TimeOfValidity = LibStrong::Type<boost::posix_time::ptime, struct TimeOfValidityTag_>;
using TemplateNameEmailSubject = StrongString<struct TemplateNameEmailSubjectTag_>;
using TemplateNameEmailBody = StrongString<struct TemplateNameEmailBodyTag_>;

LibHermes::Email::Email::Type get_libhermes_email_type(const Notification::NotifiedEvent _notified_event)
{
    switch (_notified_event)
    {
        case Notification::NotifiedEvent::created:
            return LibHermes::Email::Email::Type{"notification_create"};
        case Notification::NotifiedEvent::updated:
            return LibHermes::Email::Email::Type{"notification_update"};
        case Notification::NotifiedEvent::transferred:
            return LibHermes::Email::Email::Type{"notification_transfer"};
        case Notification::NotifiedEvent::renewed:
            return LibHermes::Email::Email::Type{"notification_renew"};
        case Notification::NotifiedEvent::deleted:
            return LibHermes::Email::Email::Type{"notification_delete"};
    }
    throw std::logic_error("value doesn't exist in Notification::NotifiedEvent");
}

TemplateNameEmailSubject get_template_name_email_subject(
        const Notification::NotifiedEvent _notified_event,
        const TemplateNames& _template_names)
{
    auto find_template_name_email_subject = [&](const auto key) -> TemplateNameEmailSubject
    {
        auto value_iter = _template_names.find(key);
        if (value_iter != _template_names.end()) {
            return TemplateNameEmailSubject{value_iter->second};
        }
        throw std::logic_error("key/value pair doesn't exist in TemplateNames");
    };

    switch (_notified_event)
    {
        case Notification::NotifiedEvent::created:
            return find_template_name_email_subject(TemplateName::template_name_email_create_subject);
        case Notification::NotifiedEvent::updated:
            return find_template_name_email_subject(TemplateName::template_name_email_update_subject);
        case Notification::NotifiedEvent::transferred:
            return find_template_name_email_subject(TemplateName::template_name_email_transfer_subject);
        case Notification::NotifiedEvent::renewed:
            return find_template_name_email_subject(TemplateName::template_name_email_renew_subject);
        case Notification::NotifiedEvent::deleted:
            return find_template_name_email_subject(TemplateName::template_name_email_delete_subject);
    }
    throw std::logic_error("value doesn't exist in Notification::NotifiedEvent");
}

TemplateNameEmailBody get_template_name_email_body(
        const Notification::NotifiedEvent _notified_event,
        const TemplateNames& _template_names)
{
    auto find_template_name_email_body = [&](const auto key) -> TemplateNameEmailBody
    {
        auto value_iter = _template_names.find(key);
        if (value_iter != _template_names.end()) {
            return TemplateNameEmailBody{value_iter->second};
        }
        throw std::logic_error("key/value pair doesn't exist in TemplateNames");
    };

    switch (_notified_event)
    {
        case Notification::NotifiedEvent::created:
            return find_template_name_email_body(TemplateName::template_name_email_create_body);
        case Notification::NotifiedEvent::updated:
            return find_template_name_email_body(TemplateName::template_name_email_update_body);
        case Notification::NotifiedEvent::transferred:
            return find_template_name_email_body(TemplateName::template_name_email_transfer_body);
        case Notification::NotifiedEvent::renewed:
            return find_template_name_email_body(TemplateName::template_name_email_renew_body);
        case Notification::NotifiedEvent::deleted:
            return find_template_name_email_body(TemplateName::template_name_email_delete_body);
    }
    throw std::logic_error("value doesn't exist in Notification::NotifiedEvent");
}

LibHermes::ReferenceType to_libhermes_reference_type(Notification::ObjectType _object_type)
{
    switch (_object_type)
    {
        case Notification::ObjectType::contact:
            return LibHermes::ReferenceType::contact;
        case Notification::ObjectType::domain:
            return LibHermes::ReferenceType::domain;
        case Notification::ObjectType::nsset:
            return LibHermes::ReferenceType::nsset;
        case Notification::ObjectType::keyset:
            return LibHermes::ReferenceType::keyset;
    }
    throw std::logic_error("unexpected value of Notification::ObjectType");
}

std::string join_map(const std::map<std::string, std::string>& _map) {
    std::vector<std::string> key_value_pairs;
    std::transform(
            _map.begin(),
            _map.end(),
            std::back_inserter(key_value_pairs),
            [](const auto& item) -> auto
            {
                return item.first + "=" + item.second;
            });

    return boost::algorithm::join(key_value_pairs, " ");
}

struct Message
{
    LibHermes::Email::Email email;
    LibHermes::ReferenceType reference_type;
    ObjectUuid object_uuid;
};

::Notification::EventOnObject make_libfred_event_on_object(
        const Notification::ObjectType& _object_type,
        const Notification::NotifiedEvent& _notified_event)
{
    return {::LibFred::object_type_from_db_handle(to_string(_object_type)),
            ::Notification::notified_event_from_db_handle(to_string(_notified_event))};
}

enum struct Result
{
    event_processed,
    no_event_processed
};

Result process_one_notification_request(
        LibFred::OperationContext& _ctx,
        const MessengerEndpoint& _messenger_endpoint,
        const MessengerArchive& _messenger_archive,
        const TemplateNames& _template_names,
        const EventAgeDaysMax& _event_age_days_max)
{
    try
    {
        const auto dbres = _ctx.get_conn().exec(
            // clang-format off
            // lock exclusively
            "WITH locked AS ("
                "SELECT change, "
                       "done_by_registrar, "
                       "historyid_post_change, "
                       "svtrid "
                  "FROM notification_queue "
                   "FOR UPDATE NOWAIT "
                 "LIMIT 1 "
            ")"
            // ...and delete (it is locked until commit and deleted immediately after)
            "DELETE "
              "FROM notification_queue AS nq "
             "USING locked "
             "WHERE nq.change = locked.change "
               "AND nq.done_by_registrar = locked.done_by_registrar "
               "AND nq.historyid_post_change = locked.historyid_post_change "
               "AND nq.svtrid = locked.svtrid "
            "RETURNING "
                "locked.*, "
                "(SELECT eot.name "
                   "FROM object_history oh "
                   "JOIN object_registry oreg USING(id) "
                   "JOIN enum_object_type eot ON oreg.type = eot.id "
                  "WHERE oh.historyid = locked.historyid_post_change) "
                     "AS object_type_, "
                "(SELECT oreg.uuid "
                   "FROM object_history oh "
                   "JOIN object_registry oreg USING(id) "
                  "WHERE oh.historyid = locked.historyid_post_change) "
                     "AS object_uuid_, "
                 "(SELECT NOW()::DATE - valid_from::DATE FROM history WHERE id = locked.historyid_post_change) AS event_age_days_");
            // clang-format on

        if (dbres.size() < 1) {
            const auto message = "no record found in notification_queue";
            LIBLOG_INFO("{}", message);
            return Result::no_event_processed;
        }

        const auto object_type = from_string<Notification::ObjectType>(static_cast<std::string>(dbres[0]["object_type_"]));
        const auto notified_event = from_string<Notification::NotifiedEvent>(static_cast<std::string>(dbres[0]["change"]));
        const auto registrar_id = RegistrarId{static_cast<unsigned long long>(dbres[0]["done_by_registrar"])};
        const auto history_id_post_change = HistoryIdPostChange{static_cast<unsigned long long>(dbres[0]["historyid_post_change"])};
        const auto svtr_id = SvtrId{static_cast<std::string>(dbres[0]["svtrid"])};
        const auto object_uuid = ObjectUuid{boost::uuids::string_generator{}(static_cast<std::string>(dbres[0]["object_uuid_"]))};
        const auto event_age_days = static_cast<unsigned int>(dbres[0]["event_age_days_"]);

        if (event_age_days > *_event_age_days_max)
        {
            LIBLOG_INFO("event older than configuration parameter 'event_age_days_max' ({})", *_event_age_days_max);
            return Result::event_processed;
        }

        const ::Notification::EventOnObject event_on_object =
                make_libfred_event_on_object(object_type, notified_event);

        const ::Notification::notification_request notification_request{
                event_on_object,
                *registrar_id,
                *history_id_post_change,
                *svtr_id};

        const auto template_name_email_subject =
                get_template_name_email_subject(notified_event, _template_names);

        const auto template_name_email_body =
                get_template_name_email_body(notified_event, _template_names);

        const std::set<std::string> recipient_email_addresses =
                gather_email_addresses(
                        _ctx,
                        notification_request.event,
                        notification_request.history_id_post_change);

        const std::map<std::string, std::string> template_parameters =
                gather_email_content(
                        _ctx,
                        notification_request);

        const auto no_change_was_done_according_to = [](const std::map<std::string, std::string>::const_iterator template_parameter_changes) {
            return template_parameter_changes->second == "0";
        };

        const auto get_template_parameter =
                [](const std::map<std::string, std::string>& template_parameters, const std::string& template_parameter_name)
                        -> std::map<std::string, std::string>::const_iterator
                {
                    const auto specific_template_parameter = template_parameters.find(template_parameter_name);
                    if (specific_template_parameter == template_parameters.end())
                    {
                        //throw ExceptionMissingChangesFlagInUpdateNotificationContent();
                        throw std::runtime_error("Missing parameter " + template_parameter_name + " in update notification content");
                    }
                    return specific_template_parameter;
        };

        // Ticket #6547 send update notifications only if there are some changes
        if (notified_event == Notification::NotifiedEvent::updated)
        {
            const auto specific_template_parameter = get_template_parameter(template_parameters, "changes");
            if (no_change_was_done_according_to(specific_template_parameter))
            {
                LIBLOG_INFO("{}", "no changes was done according to template parameter 'changes' value");
                return Result::event_processed;
            }
        }

        LIBLOG_INFO(
                "event=\"{}\" object_type=\"{}\" done_by_registrar=\"{}\" history_id_post_change=\"{}\" svtrid=\"{}\" "
                "completed - transaction not yet comitted",
                to_string(notified_event),
                to_string(object_type),
                *registrar_id,
                *history_id_post_change,
                *svtr_id);

        LibHermes::Connection<LibHermes::Service::EmailMessenger> connection{
            LibHermes::Connection<LibHermes::Service::EmailMessenger>::ConnectionString{
                *_messenger_endpoint}};

        for (const auto& recipient_email_address : recipient_email_addresses)
        {

            auto email =
                    LibHermes::Email::make_minimal_email(
                            LibHermes::Email::Email::Recipient{recipient_email_address},
                            LibHermes::Email::Email::SubjectTemplate{*template_name_email_subject},
                            LibHermes::Email::Email::BodyTemplate{*template_name_email_body});
            email.type = get_libhermes_email_type(notified_event);

            std::transform(
                    template_parameters.begin(),
                    template_parameters.end(),
                    std::inserter(email.context, email.context.end()),
                    [](const auto& item) -> auto
                    {
                        return std::make_pair(LibHermes::StructKey{item.first}, LibHermes::StructValue{item.second});
                    });

            const Message message{
                    email,
                    to_libhermes_reference_type(object_type),
                    object_uuid};

            try
            {
                    LibHermes::Email::send(
                            connection,
                            message.email,
                            LibHermes::Email::Archive{*_messenger_archive},
                            {LibHermes::Reference{
                                    message.reference_type,
                                    boost::uuids::to_string(*message.object_uuid)}});
            }
            catch (const LibHermes::Email::SendFailed& e)
            {
                LIBLOG_WARNING("gRPC exception caught while sending email about event of object with uuid {}: gRPC error code: {}, error message: {}",
                        boost::uuids::to_string(*message.object_uuid), e.error_code(), e.error_message());
                continue;
            }
            catch (const std::exception& e)
            {
                LIBLOG_WARNING("std::exception caught while sending email about event of object with uuid {}: {}", boost::uuids::to_string(*message.object_uuid), e.what());

                LIBLOG_ERROR(
                        "send_object_event_notification_emails_impl: failed to send email "
                        "event: {} object_type: {} event_done_by_registrar: {} history_id_post_change: {} "
                        "failed_recipient: {} email_template_subject: {} email_template_body: {} email_parameters: {}",
                        to_string(notified_event),
                        to_string(object_type),
                        *registrar_id,
                        *history_id_post_change,
                        recipient_email_address,
                        *template_name_email_subject,
                        *template_name_email_body,
                        join_map(template_parameters));
            }
            catch (...)
            {
                LIBLOG_WARNING("exception caught while sending email about event of object with uuid {}", boost::uuids::to_string(*message.object_uuid));

                LIBLOG_ERROR(
                        "send_object_event_notification_emails_impl: failed to send email "
                        "event: {} object_type: {} event_done_by_registrar: {} history_id_post_change: {} "
                        "failed_recipient: {} email_template_subject: {} email_template_body: {} email_parameters: {}",
                        to_string(notified_event),
                        to_string(object_type),
                        *registrar_id,
                        *history_id_post_change,
                        recipient_email_address,
                        *template_name_email_subject,
                        *template_name_email_body,
                        join_map(template_parameters));
            }
        }

    }
    catch (const std::exception& e)
    {
        const std::string what_string(e.what());
        // legacy solution, no means to identify this error better way in legacy db layer implemantation
        if (what_string.find("could not obtain lock on row in relation \"notification_queue\"") != std::string::npos)
        {
            throw FailedToLockRequest{e.what()};
        }
        throw;
    }
    return Result::event_processed;
}

} // namespace Fred::Notify::ObjectEvents::Impl::{anonymous}

void notify_object_events(
        const MessengerEndpoint& _messenger_endpoint,
        const MessengerArchive& _messenger_archive,
        const TemplateNames& _template_names,
        const EventAgeDaysMax& _event_age_days_max)
{
    while (true)
    {
        LibFred::OperationContextCreator ctx;
        try
        {
            const auto result = process_one_notification_request(ctx, _messenger_endpoint, _messenger_archive, _template_names, _event_age_days_max);
            if (result == Result::no_event_processed)
            {
                ctx.commit_transaction();
                break;
            }
        }
        catch (const FailedToLockRequest& e)
        {
            const auto error = "unable to get data from notification queue - another process is running";
            LIBLOG_ERROR("{}", error);
            throw std::runtime_error(error);
        }
        ctx.commit_transaction();
    }
}

} // namespace Fred::Notify::ObjectEvents::Impl
} // namespace Fred::Notify::ObjectEvents
} // namespace Fred::Notify
} // namespace Fred
