/*
 * Copyright (C) 2021-2022  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/notify_object_state_changes.hh"

#include "src/cfg.hh"
#include "src/impl/exception.hh"
#include "src/impl/strong_type.hh"
#include "src/impl/unwrap_uuid_from_psql.hh"

#include "src/util/util.hh"

#include "libhermes/libhermes.hh"

#include "liblog/liblog.hh"

#include "libpg/pg_rw_transaction.hh"
#include "libpg/query.hh"

#include "libstrong/type.hh"

#include <boost/functional/hash.hpp>
#include <boost/uuid/string_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <vector>

namespace Fred {
namespace Notify {
namespace ObjectStateChanges {
namespace Impl {

namespace {

/**
 * Helps to implement from_db_handle function using to_db_handle conversion function.
 * @tparam S type of source value
 * @tparam D type of destination value
 * @tparam items number of all possible destination values
 * @param _src source value
 * @param _set_of_results C array of all possible destination values
 * @param _forward_transformation function which transforms D value into S value
 * @return matching counterpart value
 * @throw std::invalid_argument in case that conversion fails
 */
template <class S, class D, unsigned items>
D inverse_transformation(const S& _src, const D (&_set_of_results)[items], S (*_forward_transformation)(D))
{
    const D* const end_of_results = _set_of_results + items;
    for (const D* result_candidate_ptr = _set_of_results; result_candidate_ptr < end_of_results; ++result_candidate_ptr)
    {
        if (_forward_transformation(*result_candidate_ptr) == _src)
        {
            return *result_candidate_ptr;
        }
    }
    throw std::invalid_argument("the value \"" + _src + "\" is invalid: conversion does not exist");
}

std::string to_string(ObjectType _value)
{
    switch (_value)
    {
        case ObjectType::contact: return "contact";
        case ObjectType::nsset: return "nsset";
        case ObjectType::domain: return "domain";
        case ObjectType::keyset: return "keyset";
    }
    throw std::logic_error("value doesn't exist in ObjectState");
}

std::string to_string(RecipientGroup _value)
{
    switch (_value)
    {
        case RecipientGroup::admin: return "admin";
        case RecipientGroup::tech: return "tech";
        case RecipientGroup::generic: return "generic";
        case RecipientGroup::additional: return "additional";
    }
    throw std::logic_error("value doesn't exist in RecipientGroup");
}

} // namespace Fred::Notify::ObjectStateChanges::Impl::{anonymous}

template <>
ObjectType from_string<ObjectType>(const std::string& _str)
{
    constexpr ObjectType possible_results[] =
            {
                ObjectType::contact,
                ObjectType::nsset,
                ObjectType::domain,
                ObjectType::keyset
            };
    return inverse_transformation(_str, possible_results, to_string);
}

template <>
RecipientGroup from_string<RecipientGroup>(const std::string& _str)
{
    constexpr RecipientGroup possible_results[] =
            {
                RecipientGroup::admin,
                RecipientGroup::tech,
                RecipientGroup::generic,
                RecipientGroup::additional
            };
    return inverse_transformation(_str, possible_results, to_string);
}

bool MailProperties::operator<(const MailProperties& _other) const
{
    return std::make_tuple(mail_template_subject, mail_template_body, mail_type) <
           std::make_tuple(_other.mail_template_subject, _other.mail_template_body, _other.mail_type);
}

decltype(auto) wrap_into_psql_representation(RecipientGroup _value)
{
    switch (_value)
    {
        case RecipientGroup::admin: return std::string{"admin"};
        case RecipientGroup::tech: return std::string{"tech"};
        case RecipientGroup::generic: return std::string{"generic"};
        case RecipientGroup::additional: return std::string{"additional"};
    }
    struct ConversionFailure : InvalidRecipientGroup
    {
        explicit ConversionFailure(std::string _msg)
            : InvalidRecipientGroup(std::move(_msg))
        {
        }
    };
    throw ConversionFailure{"unexpected value of RecipientGroup"};
}

Flag unwrap_from_psql_representation(const Flag&, const char* _str)
{
    try
    {
        return Flag{std::string{_str}};
    }
    catch (const std::runtime_error&)
    {
        struct ConversionFailure : InvalidFlag
        {
            explicit ConversionFailure(std::string _msg)
                : InvalidFlag(std::move(_msg))
            {
            }
        };
        throw ConversionFailure{"unable to convert string \"" + std::string{_str} + "\" to Flag"};
    }
}

decltype(auto) wrap_into_psql_representation(ObjectType _value)
{
    switch (_value)
    {
        case ObjectType::contact: return std::string{"contact"};
        case ObjectType::nsset: return std::string{"nsset"};
        case ObjectType::domain: return std::string{"domain"};
        case ObjectType::keyset: return std::string{"keyset"};
    }
    struct ConversionFailure : InvalidObjectType
    {
        explicit ConversionFailure(std::string _msg)
            : InvalidObjectType(std::move(_msg))
        {
        }
    };
    throw ConversionFailure{"unexpected value of ObjectType"};
}


ObjectType unwrap_from_psql_representation(const ObjectType&, const char* _str)
{
    try
    {
        return from_string<ObjectType>(_str);
    }
    catch (const std::runtime_error&)
    {
        struct ConversionFailure : InvalidObjectType
        {
            explicit ConversionFailure(std::string _msg)
                : InvalidObjectType(std::move(_msg))
            {
            }
        };
        throw ConversionFailure{"unable to convert string \"" + std::string{_str} + "\" to ObjectType"};
    }
}

RecipientGroup unwrap_from_psql_representation(const RecipientGroup&, const char* _str)
{
    try
    {
        return from_string<RecipientGroup>(_str);
    }
    catch (const std::runtime_error&)
    {
        struct ConversionFailure : InvalidRecipientGroup
        {
            explicit ConversionFailure(std::string _msg)
                : InvalidRecipientGroup(std::move(_msg))
            {
            }
        };
        throw ConversionFailure{"unable to convert string \"" + std::string{_str} + "\" to RecipientGroup"};
    }
}

namespace {

using StateId = LibStrong::Type<int, struct StateIdTag_>;
using ObjectId = LibStrong::Type<int, struct ObjectIdTag_>;
using ObjectUuid = LibStrong::Type<boost::uuids::uuid, struct ObjectUuidTag_>;

using ContactId = LibStrong::Type<int, struct ContactIdTag_>;
using DomainId = LibStrong::Type<int, struct DomainIdTag_>;
using NssetId = LibStrong::Type<int, struct NssetIdTag_>;
using KeysetId = LibStrong::Type<int, struct KeysetIdTag_>;

using Fqdn = StrongString<struct FqdnTag_>;
using Email = StrongString<struct EmailTag_>;

std::vector<Email> get_domain_admin_emails_history(
        const LibPg::PgTransaction& _tx,
        const DomainId& _domain_id)
{
    std::vector<Email> emails;

    const auto query_registrant = LibPg::make_query() <<
            // clang-format off
            "SELECT" << LibPg::item<Email>() << "ch.email "
              "FROM object_registry dor "
              "JOIN domain_history dh ON dh.historyid = dor.historyid "
              "JOIN object_registry cor ON cor.id = dh.registrant "
              "JOIN contact_history ch ON ch.historyid = cor.historyid "
             "WHERE dor.id = " << LibPg::parameter<DomainId>().as_integer();

    const auto dbres_registrant = exec(_tx, query_registrant, {_domain_id});

    for (const auto& row : dbres_registrant)
    {
        emails.push_back(row.get<Email>());
    }

    const auto query_related = LibPg::make_query() <<
            // clang-format off
            "SELECT" << LibPg::item<Email>() << "ch.email "
              "FROM contact_history ch "
              "JOIN object_registry cor ON cor.historyid = ch.historyid "
              "JOIN domain_contact_map_history dcm ON dcm.contactid = cor.id "
              "JOIN object_registry dor ON dor.historyid = dcm.historyid "
             "WHERE dor.id = " << LibPg::parameter<DomainId>().as_integer();
            // clang-format on

    const auto dbres_related = exec(_tx, query_related, {_domain_id});

    for (const auto& row : dbres_related)
    {
        emails.push_back(row.get<Email>());
    }
    return emails;
}

std::vector<Email> get_domain_tech_emails_history(
        const LibPg::PgTransaction& _tx,
        const DomainId& _domain_id)
{
    const auto query = LibPg::make_query() <<
            // clang-format off
            "SELECT" << LibPg::item<Email>() << "ch.email "
              "FROM contact_history ch "
              "JOIN object_registry cor ON cor.historyid = ch.historyid "
              "JOIN nsset_contact_map_history ncm ON ncm.contactid = cor.id "
              "JOIN object_registry nor ON nor.historyid = ncm.historyid "
              "JOIN domain_history dh ON dh.nsset = nor.id "
              "JOIN object_registry dor ON dor.historyid = dh.historyid "
             "WHERE dor.id = " << LibPg::parameter<DomainId>().as_integer();
            // clang-format on

    const auto dbres = exec(_tx, query, {_domain_id});

    std::vector<Email> emails;

    for (const auto& row : dbres)
    {
        emails.push_back(row.get<Email>());
    }
    return emails;
}

// ticket #14873
std::vector<Email> get_recipients_domain_additional(
        const LibPg::PgTransaction& _tx,
        const DomainId& _domain_id,
        const StateId& _state_id)
{
    // select unnotified (state_id IS NULL) emails inserted
    // into notify_outzone_unguarded_domain_additional_email
    // between domain expiration date (exdate) and current object state (valid_from)
    const auto query = LibPg::make_query() <<
            // clang-format off
            "SELECT" << LibPg::item<Email>() << "n.email "
              "FROM notify_outzone_unguarded_domain_additional_email n "
              "JOIN object_state os ON os.object_id = n.domain_id "
              "JOIN domain d ON d.id = n.domain_id "
            "WHERE os.id = " << LibPg::parameter<StateId>().as_integer() << " "
              "AND n.domain_id = " << LibPg::parameter<DomainId>().as_integer() << " "
              "AND n.state_id IS NULL "
              "AND n.crdate BETWEEN d.exdate AND os.valid_from";
            // clang-format on

    const auto dbres = exec(_tx, query, {_domain_id, _state_id});

    std::vector<Email> emails;

    for (const auto& row : dbres)
    {
        emails.push_back(row.get<Email>());
    }
    return emails;
}

std::vector<Email> get_nsset_tech_emails_history(
        const LibPg::PgTransaction& _tx,
        const NssetId& _nsset_id)
{
    const auto query = LibPg::make_query() <<
            // clang-format off
            "SELECT" << LibPg::item<Email>() << "ch.email "
              "FROM contact_history ch "
              "JOIN object_registry cor ON cor.historyid = ch.historyid "
              "JOIN nsset_contact_map_history ncmap ON ncmap.contactid = ch.id "
              "JOIN nsset_history nh ON ncmap.historyid = nh.historyid "
              "JOIN object_registry nobr ON nobr.historyid = nh.historyid "
             "WHERE nobr.id = " << LibPg::parameter<NssetId>().as_integer();
            // clang-format on

    const auto dbres = exec(_tx, query, {_nsset_id});

    std::vector<Email> emails;

    for (const auto& row : dbres)
    {
        emails.push_back(row.get<Email>());
    }
    return emails;
}

std::vector<Email> get_keyset_tech_emails_history(
        const LibPg::PgTransaction& _tx,
        const KeysetId& _keyset_id)
{
    const auto query = LibPg::make_query() <<
            // clang-format off
            "SELECT" << LibPg::item<Email>() << "ch.email "
              "FROM contact_history ch "
              "JOIN object_registry cor ON cor.historyid = ch.historyid "
              "JOIN keyset_contact_map_history kcmap ON kcmap.contactid = ch.id "
              "JOIN keyset_history kh ON kcmap.historyid = kh.historyid "
              "JOIN object_registry kobr ON kobr.historyid = kh.historyid "
             "WHERE kobr.id = " << LibPg::parameter<KeysetId>().as_integer();
            // clang-format on
 
    const auto dbres = exec(_tx, query, {_keyset_id});

    std::vector<Email> emails;

    for (const auto& row : dbres)
    {
        emails.push_back(row.get<Email>());
    }
    return emails;
}

std::vector<Email> get_contact_emails_history(
        const LibPg::PgTransaction& _tx,
        const ContactId& _contact_id)
{
    const auto query = LibPg::make_query() <<
            // clang-format off`
            "SELECT" << LibPg::item<Email>() << "ch.email "
              "FROM contact_history ch "
              "JOIN object_registry cor ON ch.historyid = cor.historyid "
             "WHERE cor.id = " << LibPg::parameter<ContactId>().as_integer();
            // clang-format on

    const auto dbres = exec(_tx, query, {_contact_id});

    std::vector<Email> emails;

    for (const auto& row : dbres)
    {
        emails.push_back(row.get<Email>());
    }
    return emails;
}

// ticket #3797
std::vector<Email> get_recipients_domain_generic(
        const Fqdn& _fqdn)
{
    if ((*_fqdn).empty())
    {
        throw std::runtime_error{"fqdn empty"};
    }
    std::vector<Email> emails;
    for (const auto& generic_email_prefix : Cfg::Options::get().notification.generic_email_prefixes)
    {
        emails.push_back(Email(generic_email_prefix + *_fqdn));
    }
    emails.push_back(Email{(*_fqdn).substr(0, (*_fqdn).find(".")) + "@" + *_fqdn});
    return emails;
}

template <typename T>
std::vector<Email> get_recipients(const LibPg::PgTransaction&, const T&, RecipientGroup);

template <>
std::vector<Email> get_recipients<ContactId>(const LibPg::PgTransaction& _tx, const ContactId& _contact_id, RecipientGroup _recipient)
{
    switch (_recipient)
    {
        case RecipientGroup::tech:
            return get_contact_emails_history(_tx, _contact_id);
        case RecipientGroup::admin:
        case RecipientGroup::generic:
        case RecipientGroup::additional:
            break;
    }
    throw std::runtime_error("invalid configuration: contact does not expect recipient group " + to_string(_recipient));
}

template <>
std::vector<Email> get_recipients<NssetId>(const LibPg::PgTransaction& _tx, const NssetId& _nsset_id, RecipientGroup _recipient)
{
    switch (_recipient)
    {
        case RecipientGroup::tech:
            return get_nsset_tech_emails_history(_tx, _nsset_id);
        case RecipientGroup::admin: 
        case RecipientGroup::generic:
        case RecipientGroup::additional:
            break;
    }
    throw std::runtime_error("invalid configuration: nsset does not expect recipient group " + to_string(_recipient));
}

template <>
std::vector<Email> get_recipients<DomainId>(const LibPg::PgTransaction& _tx, const DomainId& _domain_id, RecipientGroup _recipient)
{
    switch (_recipient)
    {
        case RecipientGroup::admin:
            return get_domain_admin_emails_history(_tx, _domain_id);
        case RecipientGroup::tech:
            return get_domain_tech_emails_history(_tx, _domain_id);
        case RecipientGroup::generic:
        case RecipientGroup::additional:
            break;
    }
    throw std::runtime_error("invalid configuration: domain does not expect recipient group " + to_string(_recipient));
}

template <>
std::vector<Email> get_recipients<KeysetId>(const LibPg::PgTransaction& _tx, const KeysetId& _keyset_id, RecipientGroup _recipient)
{
    switch (_recipient)
    {
        case RecipientGroup::tech:
            return get_keyset_tech_emails_history(_tx, _keyset_id);
        case RecipientGroup::admin:
        case RecipientGroup::generic:
        case RecipientGroup::additional:
            break;
    }
    throw std::runtime_error("invalid configuration: keyset does not expect recipient group " + to_string(_recipient));
}

LibHermes::ReferenceType to_libhermes_reference_type(ObjectType _object_type)
{
    switch (_object_type)
    {
        case ObjectType::contact:
            return LibHermes::ReferenceType::contact;
        case ObjectType::domain:
            return LibHermes::ReferenceType::domain;
        case ObjectType::nsset:
            return LibHermes::ReferenceType::nsset;
        case ObjectType::keyset:
            return LibHermes::ReferenceType::keyset;
    }
    throw std::logic_error("unexpected value of ObjectType");
}

struct Message
{
    LibHermes::Email::Email email;
    StateId state_id;
    ObjectUuid object_uuid;

    ObjectId object_id;
    ObjectType object_type;
    RecipientGroup recipient_group;
};

void create_configuration_table(
        const LibPg::PgRwTransaction& _tx,
        const Notifications& _notifications)
{
    exec(_tx, LibPg::make_query() <<
            "CREATE TEMPORARY TABLE tmp_configuration ( "
            "object_state_id INTEGER NOT NULL, "
            "recipient_group notification.RECIPIENT_GROUP NOT NULL, "
            "object_type INTEGER NOT NULL)");

    for (const auto& notification : _notifications)
    {
        const auto object_type = std::get<ObjectType>(notification.first);
        const auto flag = std::get<Flag>(notification.first);

        for (const auto& mail_properties_with_recipients : notification.second)
        {
            const auto recipients_groups = mail_properties_with_recipients.second;

            for (const auto& recipient_group : recipients_groups)
            {
                LIBLOG_INFO("{} | {} | {}", to_string(object_type), *flag, to_string(recipient_group));

                exec(_tx, LibPg::make_query() <<
                        "INSERT INTO tmp_configuration (object_state_id, recipient_group, object_type) "
                        "VALUES (" <<
                        "(SELECT id FROM enum_object_states WHERE name = " << LibPg::parameter<Flag>().as_text() << "), " <<
                        LibPg::parameter<RecipientGroup>().as_type("notification.RECIPIENT_GROUP") << ", " <<
                        "(SELECT id FROM enum_object_type WHERE name = " << LibPg::parameter<ObjectType>().as_text() << "))",
                        {flag, recipient_group, object_type});
            }
        }
    }
}

std::vector<Message> get_messages(
        const Notifications& _notifications,
        const NotificationsMax& _notifications_max)
{
    auto tx = Util::make_rw_transaction();

    create_configuration_table(tx, _notifications);

    const auto query = LibPg::make_query() <<
            // clang-format off
            "SELECT" << LibPg::item<StateId>() << "os.id" <<
                         LibPg::item<ObjectId>() << "oreg.id" <<
                         LibPg::item<ObjectUuid>() << "oreg.uuid" <<
                         LibPg::item<ObjectType>() << "eot.name" <<
                         LibPg::item<Flag>() << "eos.name" <<
                         LibPg::item<Fqdn>() << "oreg.name" <<
                         LibPg::item<RecipientGroup>() << "tc.recipient_group "
              " FROM object_state os "
              "JOIN enum_object_states eos ON eos.id = os.state_id "
              "JOIN object_registry oreg ON oreg.id = os.object_id "
              "JOIN enum_object_type eot ON eot.id = oreg.type "
              "JOIN tmp_configuration tc ON (tc.object_state_id = os.id AND tc.object_type = oreg.type) "
             "WHERE os.valid_to IS NULL "
               "AND NOT EXISTS ("
                      "SELECT 0 FROM notification.object_state_changes nosc "
                       "WHERE nosc.object_state_id = os.id "
                         "AND nosc.message_type = 'email'::notification.MESSAGE_TYPE "
                         "AND nosc.message_ident IS NOT NULL) "
             "ORDER BY os.id "
             "LIMIT " << LibPg::parameter<NotificationsMax>().as_integer();
            // clang-format on

    const auto dbres = exec(tx, query, {_notifications_max});

    std::vector<Message> messages;

    for (const auto& row : dbres)
    {
        const auto state_id = row.get<StateId>();
        const auto object_id = row.get<ObjectId>();
        const auto object_uuid = row.get<ObjectUuid>();
        const auto object_type = row.get<ObjectType>();
        const auto flag = row.get<Flag>();
        const auto fqdn = row.get<Fqdn>();
        const auto recipient_group = row.get<RecipientGroup>();

        const auto notifications = _notifications.find(std::make_tuple(object_type, flag));

        if (notifications == _notifications.end())
        {
            continue;
        }

        for (const auto& mail_properties_with_recipients : notifications->second)
        {
            const auto mail_properties = mail_properties_with_recipients.first;
            const auto recipients_groups = mail_properties_with_recipients.second;

            const auto recipients =
                    [&](ObjectType object_type, const RecipientGroup& recipient_group){
                        switch (object_type)
                        {
                            case ObjectType::contact:
                                    return get_recipients(tx, ContactId{*object_id}, recipient_group);
                            case ObjectType::nsset:
                                    return get_recipients(tx, NssetId{*object_id}, recipient_group);
                            case ObjectType::domain:
                                    switch (recipient_group)
                                    {
                                        case RecipientGroup::additional:
                                            return get_recipients_domain_additional(tx, DomainId{*object_id}, state_id);
                                        case RecipientGroup::generic:
                                            return get_recipients_domain_generic(fqdn);
                                        default:
                                            return get_recipients(tx, DomainId{*object_id}, recipient_group);
                                    }
                            case ObjectType::keyset:
                                    return get_recipients(tx, KeysetId{*object_id}, recipient_group);
                        }
                        throw std::logic_error("unexpected ObjectType");
                    }(object_type, recipient_group);

            if (recipients.empty())
            {
                LIBLOG_WARNING("object of type {} has no recipients", to_string(object_type));
                continue;
            }

            for (const auto& recipient : recipients)
            {
                auto email =
                        LibHermes::Email::make_minimal_email(
                                LibHermes::Email::Email::Recipient{boost::algorithm::trim_copy(*recipient)},
                                LibHermes::Email::Email::SubjectTemplate{*mail_properties.mail_template_subject},
                                LibHermes::Email::Email::BodyTemplate{*mail_properties.mail_template_body});
                email.type = LibHermes::Email::Email::Type{*mail_properties.mail_type};

                messages.push_back(Message{std::move(email), state_id, object_uuid, object_id, object_type, recipient_group});
            }
        }
    }

    commit(std::move(tx));
    return messages;
}

void save_confirmation(
        const LibPg::PgRwTransaction& _tx_rw,
        const StateId& _state_id,
        const RecipientGroup& _recipient_group,
        const LibHermes::Email::EmailUid& _email_uid)
{
    const auto query = LibPg::make_query() <<
           "INSERT INTO notification.object_state_changes "
           "(state_id, recipient_group, message_type, message_ident) "
           "VALUES (" << LibPg::parameter<StateId>().as_integer() << ", " <<
                         LibPg::parameter<RecipientGroup>().as_type("notification.RECIPIENT_GROUP") << ", " <<
                         "'email'::notification.MESSAGE_TYPE, " <<
                         LibPg::parameter<LibHermes::Email::EmailUid>().as_text() << ")";
    exec(_tx_rw, query, {_state_id, _recipient_group, _email_uid});
}

void save_domain_additional_email_state(
        const LibPg::PgRwTransaction& _tx_rw,
        const StateId& _state_id,
        const DomainId& _domain_id,
        const Email& _email)
{
    const auto query = LibPg::make_query() <<
            // clang-format off
            "UPDATE notify_outzone_unguarded_domain_additional_email "
               "SET state_id = " << LibPg::parameter<StateId>().as_integer() << " "
             "WHERE domain_id = " << LibPg::parameter<DomainId>().as_integer() << " "
               "AND email = " << LibPg::parameter<Email>().as_text() << " "
               "AND state_id IS NULL";
            // clang-format on
    exec(_tx_rw, query, {_state_id, _domain_id, _email});
}

} // namespace Fred::Notify::ObjectStateChanges::Impl::{anonymous}

void notify_object_state_changes(
        const Notifications& _notifications,
        const NotificationsMax& _notifications_max,
        const MessengerEndpoint& _messenger_endpoint,
        const MessengerArchive& _messenger_archive)
{
    try
    {
        const auto messages = get_messages(_notifications, _notifications_max);

        LibHermes::Connection<LibHermes::Service::EmailMessenger> connection{
            LibHermes::Connection<LibHermes::Service::EmailMessenger>::ConnectionString{
                *_messenger_endpoint}};

        auto conn_ptr = std::make_unique<LibPg::PgConnection>(Util::make_rw_connection());

        for (const auto& message : messages)
        {
            LibHermes::Email::EmailUid email_uid;
            try
            {
                email_uid =
                        LibHermes::Email::send(
                                connection,
                                message.email,
                                LibHermes::Email::Archive{*_messenger_archive},
                                {LibHermes::Reference{
                                        to_libhermes_reference_type(message.object_type),
                                        boost::uuids::to_string(*message.object_uuid)}});
            }
            catch (const LibHermes::Email::SendFailed& e)
            {
                LIBLOG_WARNING("gRPC exception caught while sending email obout object with uuid {}: gRPC error code: {}, error message: {}",
                        boost::uuids::to_string(*message.object_uuid), e.error_code(), e.error_message());
                continue;
            }
            catch (const std::exception& e)
            {
                LIBLOG_WARNING("std::exception caught while sending email about object with uuid {}: {}", boost::uuids::to_string(*message.object_uuid), e.what());
                continue;
            }
            catch (...)
            {
                LIBLOG_WARNING("exception caught while sending email about object with uuid {}", boost::uuids::to_string(*message.object_uuid));
                continue;
            }

            auto tx = LibPg::PgRwTransaction{std::move(*(conn_ptr.release()))};
            save_confirmation(tx, message.state_id, message.recipient_group, email_uid);
            if (message.object_type == ObjectType::domain &&
                message.recipient_group == RecipientGroup::additional)
            {
                save_domain_additional_email_state(tx, message.state_id, DomainId{*message.object_id}, Email{*message.email.recipient});
            }
            conn_ptr = std::make_unique<LibPg::PgConnection>(commit(std::move(tx)));
        }
    }
    catch (const LibPg::ExecFailure& e)
    {
        LIBLOG_WARNING("SQL exec failure: {}", e.what());
        throw;
    }
    catch (const LibPg::ConnectFailure& e)
    {
        LIBLOG_WARNING("DB connect failure: {}", e.what());
        throw;
    }
    catch (const std::exception& e)
    {
        LIBLOG_WARNING("std::exception caught: {}", e.what());
        throw;
    }
}

} // namespace Fred::Notify::ObjectStateChanges::Impl
} // namespace Fred::Notify::ObjectStateChanges
} // namespace Fred::Notify
} // namespace Fred
